/* $Copyright$
 *  Copyright (C) Fujitsu Siemens Computers GmbH 2001
 *  All rights reserved
 */
#ident "$Header$"
#ifndef WITHOUT_INCLUDES
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE_SUB
#  define MODULE
#endif
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/pci.h>
#include <asm/io.h>            	/* inb */
#include <asm/uaccess.h>		/* put_user */
#if LINUX_VERSION_CODE < 0x020400
#define vm_pgoff                        vm_offset
#endif
#endif
#ident "$Header: $"
#ident "$Header$"
typedef struct ihpci_ioctl_struct
{
        int NoOfAsics;
        unsigned long physShmemAdr[16];
        size_t shmemSize[16];
} ihpci_ioctl_t;
typedef struct __IoctlSlotInformation
{
        unsigned long PhysicalSlotNumber;
        unsigned long GreenLEDBehavior;
        unsigned long AmberLEDBehavior;
        unsigned long PowerBehavior;
        unsigned long PRSNT12;
        unsigned long SlotSwitchStatus;
} DATA_SLOT_INFO, *P_DATA_SLOT_INFO;
typedef DATA_SLOT_INFO IOCTL_DATA_SLOT_INFO, *P_IOCTL_DATA_SLOT_INFO;
typedef struct __IoctlSystemInformation
{
        unsigned long NumberHotPlugSlot;
        unsigned long FirstHotPlugSlot;
        unsigned long LastHotPlugSlot;
} DATA_SYSTEM_INFO, *P_DATA_SYSTEM_INFO;
typedef DATA_SYSTEM_INFO IOCTL_DATA_SYSTEM_INFO, *P_IOCTL_DATA_SYSTEM_INFO;
EXPORT_NO_SYMBOLS;
static const int SUCCESS = 0;
static const char *IHPCI_DEV = "ihpci";
static const unsigned int PCI_VENDORID_CPQ = 0xE11;
static const unsigned int PCI_DEVICE_ID_HPCTRL = 0xA0F7;
static const unsigned int HOTPLUG_BASEADDR = 0x0010;
        static short ihpciDebug = 0;
        MODULE_PARM(ihpciDebug, "h");
static int ihpci_major = 0;
static int g_deviceOpen = 0;
static int g_noOfAsics = -1;
static struct pci_dev *g_dev[16+1];
static u32 g_baseAddr[16];
static int g_deviceMap[16];
static int ihpci_ioctl(
                                struct inode *inode,
                                struct file *file,
                                unsigned int cmd,
                                unsigned long arg )
{
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "-->ihpci_ioctl(0x%x)\n",cmd);
        if (_IOC_TYPE(cmd) != 'i') {
                if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "  type != magic\n");
                return -EINVAL;
        }
        if (_IOC_NR(cmd) > 1) {
                if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "  nr > maxnr\n");
                return -EINVAL;
        }
        switch (cmd) {
        case _IOR ('i', 0, ihpci_ioctl_t):
                {
                        int i;
                        ihpci_ioctl_t *asicinfo = (ihpci_ioctl_t *)arg;
                        unsigned long flags;
                        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" " IHPCI_IOC_GET_ASIC_INFO\n");
                        save_flags(flags);
                        cli();
                        put_user(g_noOfAsics, (int *)&(asicinfo->NoOfAsics));
                        for (i=0; i < g_noOfAsics ; i++) {
                                put_user(g_baseAddr[i], (unsigned long *)&(asicinfo->physShmemAdr[i]));
                                put_user(512,(size_t *)&(asicinfo->shmemSize[i]));
                        }
                        restore_flags(flags);
                }
                break;
        default:
                if (ihpciDebug) printk(KERN_DEBUG "ihpci :" " ioctl-cmd=%x not valid\n",cmd);
                return -EINVAL;
        }
        return SUCCESS;
}
static int ihpci_open(struct inode *inode, struct file *file)
{
    if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "-->ihpci_open: (%p,%p)\n",inode, file);
        if (g_deviceOpen > 0) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "ihpci_open: device is busy\n");
        return -EBUSY;
        }
    if(g_noOfAsics < 0) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "ihpci_open: device not initialized\n");
        return -ENODEV;
    }
        g_deviceOpen++;
    MOD_INC_USE_COUNT;
    return SUCCESS;
}
static int ihpci_release(struct inode *inode, struct file *file)
{
    if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "-->ihpci_release: (%p,%p)\n",inode, file);
        g_deviceOpen--;
    MOD_DEC_USE_COUNT;
    return SUCCESS;
}
static int ihpciGetBaseAddrIndex(unsigned long off)
{
        int i;
        for (i = 0; i < g_noOfAsics; i++) {
                if (g_baseAddr[i] == off)
                        break;
        }
        if (i >= g_noOfAsics) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "ihpci_mmap: offset not valid\n");
        return -1;
        }
        return i;
}
static void ihpci_vma_unmap(
                struct vm_area_struct *vma,
                unsigned long addr,
                size_t len)
{
        int i;
        unsigned long off = vma->vm_pgoff;
    if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "-->ihpci_vma_unmap: off=0x%lx\n",off);
        i = ihpciGetBaseAddrIndex(off);
        if (i < 0) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "ihpci_vma_unmap: offset not valid\n");
        return ;
        }
        g_deviceMap[i]--;
    if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "-->ihpci_vma_unmap\n");
}
static void ihpci_vma_open(struct vm_area_struct *area) {
    if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "-->ihpci_vma_open\n");
    MOD_INC_USE_COUNT;
}
static void ihpci_vma_close(struct vm_area_struct *area) {
    if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "-->ihpci_vma_close\n");
    ihpci_vma_unmap(area, area->vm_start, (size_t)(area->vm_end - area->vm_start));
    MOD_DEC_USE_COUNT;
}
static struct vm_operations_struct ihpci_vm_ops = {
    ihpci_vma_open,
    ihpci_vma_close,
};
static int ihpci_mmap(struct file *file, struct vm_area_struct *vma)
{
    int err;
        int i;
    unsigned long virt_addr = vma->vm_start;
    unsigned long vsize = vma->vm_end - vma->vm_start;
    unsigned long off = vma->vm_pgoff;
    if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "-->ihpci_mmap: virt=0x%lx vsize=0x%lx off=0x%lx\n",virt_addr, vsize, off);
        i = ihpciGetBaseAddrIndex(off);
        if (i < 0) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "ihpci_mmap: offset not valid\n");
        return -EINVAL;
        }
        if (g_deviceMap[i] != -1) {
                if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "ihpci_mmap in use\n");
                return -ENXIO;
        }
    err = remap_page_range(
            virt_addr,
            off,
            vsize,
            vma->vm_page_prot);
    if (err) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "ihpci_mmap: remap_page_range failed: %d\n",err);
        return -EINVAL;
    }
    if (vma->vm_ops) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "ihpci_mmap: vma->vm_ops != NULL\n");
        return -EINVAL;
    }
    vma->vm_ops = &ihpci_vm_ops;
    MOD_INC_USE_COUNT;
        g_deviceMap[i]++;
    if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "<--ihpci_mmap\n");
    return SUCCESS;
}
static struct file_operations ihpci_fops =
{
    .ioctl = ihpci_ioctl,
    .mmap = ihpci_mmap,
    .open = ihpci_open,
    .release = ihpci_release
};
int init_module(void)
{
        int cnt;
        struct pci_dev *dev;
        g_noOfAsics = -1;
    if (! pci_present()) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "init_module: pcibios not present\n");
        return -ENODEV;
    }
        for ( cnt = 0; cnt < 16; cnt++)
                g_deviceMap[cnt] = -1;
        cnt = 0;
        dev = NULL;
        g_dev[cnt] = NULL;
        while ( (cnt < 16)
                        && (dev = pci_find_device( PCI_VENDORID_CPQ, PCI_DEVICE_ID_HPCTRL, dev) ) != NULL) {
                pci_read_config_dword(dev, HOTPLUG_BASEADDR, &(g_baseAddr[cnt]));
                g_baseAddr[cnt] &= 0xFFFFFFF0;
                g_dev[cnt] = dev;
                cnt++;
                g_dev[cnt] = NULL;
        }
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "init_module: noOfAsics=0x%x\n",cnt);
    if (cnt == 0) {
        if (ihpciDebug) printk(KERN_DEBUG "ihpci :" "init_module: ihpci not present!\n");
        return -ENODEV;
    }
        if ((ihpci_major=register_chrdev(ihpci_major, IHPCI_DEV, &ihpci_fops)) < 0) {
        printk(KERN_ERR "ihpci: unable to get major %d\n", ihpci_major);
        return -EIO;
        }
        g_noOfAsics = cnt;
    return SUCCESS;
}
void cleanup_module(void)
{
    int err;
        int i;
    err = unregister_chrdev(ihpci_major, IHPCI_DEV);
    if (err < 0) {
        printk(KERN_ERR "ihpci: unregister_chrdev(%i) failed!\n", ihpci_major);
    }
        for ( i = 0; i < 16; i++)
                g_deviceMap[i] = -1;
        g_deviceOpen = 0;
    g_noOfAsics = -1;
}
